# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import requests
from subprocess import run

from ..environment import BaseEnvironment
from ...general import _pkg_root
try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath


class Geant4Environment(BaseEnvironment):
    _read_only_paths = {'bdsim': 0, 'geant4': 0}

    def __init__(self):
        super().__init__()
        self._in_constructor = True
        self._geant4 = None
        self._bdsim = None
        self._in_constructor = False
        self._geant4_sourced = False
        self._bdsim_sourced = False
        cmd = run(['which', 'geant4-config'], capture_output=True)
        if cmd.returncode == 0:
            path = FsPath(cmd.stdout.decode().strip())
            if path.exists():
                self._geant4 = path
                self._geant4_sourced = True
        cmd = run(['which', 'bdsim'], capture_output=True)
        if cmd.returncode == 0:
            path = FsPath(cmd.stdout.decode().strip())
            if path.exists():
                self._bdsim = path
                self._bdsim_sourced = True

    @property
    def compiled(self):
        if self.geant4 is None or self.bdsim is None:
            return False
        so = list((self.data_dir).glob('g4interface.*so'))
        if len(so) != 1 or not so[0].exists():
            return False
        return self._geant4_sourced and self._bdsim_sourced

    def compile(self, verbose=True):
        # Check all dependencies
        self.assert_make_installed(verbose=verbose)
        self.assert_gcc_installed(verbose=verbose)
        if self.geant4 is None:
            raise RuntimeError("Could not find Geant4 installation! Please install Geant4.")
        if self.bdsim is None:
            raise RuntimeError("Could not find BDSIM installation! Please install BDSIM.")
        try:
            import pybind11 # noqa F401
        except (ModuleNotFoundError, ImportError):
            pybind_repo = 'https://github.com/pybind/pybind11.git'
            try:
                not_found = requests.get(pybind_repo).status_code == 404
            except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError):
                not_found = True
            if not_found:
                raise RuntimeError("Could not find pybind11! Please pip install "
                                  f"pybind11, or make sure {pybind_repo} is accessible.")

        # Copy the C source files to a temporary directory and compile it
        dest = (self.temp_dir / 'xcoll_bdsim').resolve()
        dest.mkdir(parents=True, exist_ok=True)
        for path in (_pkg_root / 'scattering_routines' / 'geant4' / 'scattering_src').glob('*'):
            if path.name == '.git' or path.name == 'docs' or path.name == 'tests' \
            or path.name == 'samples':
                continue
            else:
                FsPath(path).copy_to(dest, method='mount')
        cwd = FsPath.cwd()
        os.chdir(dest)

        # Compile
        cmd = run(['cmake', '-S', '.', '-B', 'build'], capture_output=True)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            os.chdir(cwd)
            raise RuntimeError(f"Failed to build Xcoll-BDSIM interface!\nError given is:\n{stderr}")
        cmd = run(['cmake', '--build', 'build'], capture_output=True)
        if cmd.returncode == 0:
            if verbose:
                print("Compiled Xcoll-BDSIM interface successfully.")
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            os.chdir(cwd)
            raise RuntimeError(f"Failed to compile Xcoll-BDSIM interface!\nError given is:\n{stderr}")
        # Collect the compiled shared library
        so = list((dest / 'build').glob('g4interface.*so'))
        if len(so) > 1:
            raise RuntimeError(f"Compiled into multiple g4interface shared libraries!")
        if len(so) == 0:
            raise RuntimeError(f"Failed Xcoll-BDSIM compilation! No shared library found in "
                             + f"{self.data_dir.as_posix()}!")
        so = FsPath(so[0])
        so.move_to(self.data_dir / so.name)
        if verbose:
            print(f"Created Xcoll-BDSIM shared library in {so}.")
        # Clean up the temporary directory
        self.temp_dir = None
        os.chdir(cwd)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .atoms import Helium, Nitrogen, Oxygen, Argon, Carbon
from .material import Material
from .database import db

# Liquid gases
LiquidHelium   = Helium.adapt(state='liquid',   density=0.146,  temperature=1.9,   pressure=None)
LiquidNitrogen = Nitrogen.adapt(state='liquid', density=0.808,  temperature=77.36, pressure=None)
LiquidOxygen   = Oxygen.adapt(state='liquid',   density=1.141,  temperature=90.20, pressure=None)
LiquidArgon    = Argon.adapt(state='liquid',    density=1.3954, temperature=87.3,  pressure=None)


# Carbon variants (need to provide name explicitly to avoid fluka_name being written as main name)
Carbon140 = Carbon.adapt(name='Carbon140', density=1.40, fluka_name='CC_1_40')
Carbon175 = Carbon.adapt(name='Carbon175', density=1.75, fluka_name='CC_1_75')
Carbon180 = Carbon.adapt(name='Carbon180', density=1.80, fluka_name='CC_1_80')
Carbon185 = Carbon.adapt(name='Carbon185', density=1.85, fluka_name='CC_1_85')
Diamond   = Carbon.adapt(name='Diamond', density=3.52, excitation_energy=88.5)
GraphiteR4550 = Carbon.adapt(name='GraphiteR4550', density=1.83, excitation_energy=78.0, fluka_name='GRAR4550')
CarbonFibreCarbon = GraphiteR4550.adapt(name='CarbonFibreCarbon', density=1.67, nuclear_radius=0.25,
                                        nuclear_elastic_slope=70.0, cross_section=[0.337, 0.232, 0, 0, 0, 0.0076e-2],
                                        hcut=0.02, fluka_name='AC150GPH')
db['CFC'] = CarbonFibreCarbon


# Metadata for database
# =====================

del Helium, Nitrogen, Oxygen, Argon, Carbon
for name, obj in list(globals().items()):  # Have to wrap in list to take a snapshot (avoid updating in-place)
    if isinstance(obj, Material) and obj.name is None:
        obj.name = name

Carbon140.info         = 'Carbon-based material. Used in LHC dump.'
Carbon175.info         = 'Carbon-based material. Used in LHC dump.'
Carbon180.info         = 'Carbon-based material. Used in LHC dump.'
Carbon185.info         = 'Carbon-based material. Used in LHC dump.'
GraphiteR4550.info     = 'Graphite material. Was a candidate for LHC collimators.'
CarbonFibreCarbon.info = 'Carbon-fibre composite. Used in LHC collimators.'

# Clean up namespace
del name, obj
del Material
del db



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .general import _pkg_root, __version__, citation

from .beam_elements import BlackAbsorber, BlackCrystal, TransparentCollimator, TransparentCrystal, \
                           EverestBlock, EverestCollimator, EverestCrystal, Geant4Collimator, Geant4Crystal, \
                           BlowUp, EmittanceMonitor, collimator_classes, crystal_classes, element_classes
from .materials import Material, CrystalMaterial, RefMaterial
from .scattering_routines.geant4 import Geant4Engine
from .colldb import CollimatorDatabase
from .interaction_record import InteractionRecord
from .rf_sweep import RFSweep, prepare_rf_sweep
from .lossmap import LossMap, MultiLossMap
from .particles_tree import ParticlesTree

from .constants import particle_states, particle_state_names, interactions, interaction_names

# Initialise Geant4 environment
from .scattering_routines.geant4.wrapper import Geant4Wrapper as _Geant4Wrapper
geant4 = _Geant4Wrapper()

# print("If you use Xcoll in your simulations, please cite us :-)")
# print(citation)



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import sys
import numpy as np
from pathlib import Path
from numbers import Number

import xobjects as xo
import xtrack as xt
import xtrack.particles.pdg as pdg

from .geometry import XcollGeometry
from ..materials import Material, CrystalMaterial
from ..interaction_record import InteractionRecord
try:
    # TODO: once xaux is in Xsuite keep only this
    from xaux import FsPath, ranID
except (ImportError, ModuleNotFoundError):
    from ..xaux import FsPath, ranID


class BaseEngine(xo.HybridClass):
    _xofields = {
        '_particle_ref':      xt.Particles._XoStruct,
        '_seed':              xo.UInt64,
        '_capacity':          xo.Int64,
        '_relative_capacity': xo.Int64,
    }

    _int32 = False
    _only_protons = False
    _element_classes = None
    _uses_input_file = False
    _num_input_files = 1
    _uses_run_folder = False

    _depends_on = [Material, CrystalMaterial, InteractionRecord, xt.RandomUniform,
                   xt.RandomExponential, xt.RandomNormal, xt.RandomRutherford,
                   xt.Drift, XcollGeometry]

    def __init__(self, **kwargs):
        if self._element_classes is None:
            raise NotImplementedError(f"{self.__class__.__name__} needs to define `_element_classes`!")
        # Initialise defaults
        self._cwd = None
        self._line = None
        self._verbose = False
        self._input_file = None
        self._element_dict = {}
        self._warning_given = False
        self._environment = None
        self._element_index = 0
        self._tracking_initialised = False
        self._deactivated_elements = {}
        kwargs.setdefault('_particle_ref', xt.Particles())
        kwargs.setdefault('_seed', 0)
        kwargs.setdefault('_capacity', 0)
        super().__init__(**kwargs)

    def __del__(self, *args, **kwargs):
        self.stop(warn=False)

    def _warn(self, error=None):
        if not self._warning_given:
            print(f"Warning: Failed to import {self.__class__.__name__} environment "
                + f"(did you compile?).\n{self.name.capitalize()} elements can be installed "
                + f"but are not trackable.", flush=True)
            if error:
                print(f"Error: {error}", flush=True)
            self._warning_given = True
        self.stop()
        if error:
            raise error

    def _print(self, *args, **kwargs):
        if self.verbose:
            kwargs.setdefault('flush', True)
            print(*args, **kwargs)


    # ==================
    # === Properties ===
    # ==================

    @property
    def environment(self):
        return self._environment

    @property
    def name(self):
        return self.__class__.__name__.replace('Engine', '').lower()

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, val):
        self._verbose = val

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, val):
        if not val is None and not isinstance(val, xt.Line):
            raise ValueError("`line` has to be an xt.Line object!")
        self._line = val

    @line.deleter
    def line(self):
        self._line = None

    @property
    def particle_ref(self):
        initial = xt.Particles().to_dict()
        current = self._particle_ref.to_dict()
        if xt.line._dicts_equal(initial, current):
            return None
        else:
            return self._particle_ref

    @particle_ref.setter
    def particle_ref(self, val):
        if val is None:
            self._particle_ref = xt.Particles()
        else:
            if isinstance(val, xt.line.LineParticleRef):
                val = val._resolved
            if not isinstance(val, xt.Particles):
                raise ValueError("`particle_ref` has to be an xt.Particles object!")
            if val._capacity > 1:
                raise ValueError("`particle_ref` has to be a single particle!")
            pdg_id = val.pdg_id[0]
            if pdg_id == 0:
                if self._only_protons:
                    pdg_id = pdg.get_pdg_id_from_name('proton')
                else:
                    raise ValueError(f"{self.__class__.__name__} allows the use of particles "
                                   + f"different than protons. Hence, `particle_ref` "
                                   + f"needs to have a valid pdg_id.")
            elif self._only_protons and pdg_id != pdg.get_pdg_id_from_name('proton'):
                raise ValueError("{self.__class__.__name__} only supports protons!")
            self._particle_ref = val
            self._particle_ref.pdg_id[0] = pdg_id

    @particle_ref.deleter
    def particle_ref(self):
        self.particle_ref = None

    @property
    def capacity(self):
        if self._capacity == 0:
            return None
        else:
            return int(self._capacity)

    @capacity.setter
    def capacity(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            raise ValueError("`capacity` has to be a positive integer!")
        self._capacity = int(val)

    @capacity.deleter
    def capacity(self):
        self.capacity = None

    @property
    def relative_capacity(self):
        if self._relative_capacity == 0:
            return None
        else:
            return int(self._relative_capacity)

    @relative_capacity.setter
    def relative_capacity(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            raise ValueError("`relative_capacity` has to be a positive integer!")
        if val <= 1:
            raise ValueError("`relative_capacity` has to be larger than 1!")
        self._relative_capacity = int(val)

    @relative_capacity.deleter
    def relative_capacity(self):
        self.relative_capacity = None

    @property
    def seed(self):
        if self._seed == 0:
            return None
        else:
            return self._seed

    @seed.setter
    def seed(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            raise ValueError("`seed` has to be a positive integer!")
        val = int(val)
        if self._int32:
            new_val = np.uint32(val)
        else:
            new_val = np.uint64(val)
        if new_val != val:
            self._print(f"Warning: type change for seed {val}. Using {new_val}.")
        self._seed = new_val

    @seed.deleter
    def seed(self):
        self.seed = None

    @property
    def input_file(self):
        if self._uses_input_file:
            return self._input_file

    @property
    def element_dict(self):
        return self._element_dict


    # ======================
    # === Public Methods ===
    # ======================

    def start(self, *, clean=True, input_file=None, **kwargs):
        if not self.environment:
            raise RuntimeError(f"{self.name.capitalize()} environment not set up! "
                             + f"Do not manually create an instance of the engine.")
        if not self.environment.initialised:
            raise RuntimeError(f"{self.name.capitalize()} environment not initialised! "
                             + f"Please set all paths in the environment before "
                             + f"starting the engine.")
        if not self.environment.compiled:
            raise RuntimeError(f"{self.name.capitalize()} interface not compiled! "
                             + f"Please compile before starting the engine.")
        if self.is_running():
            self._print("Engine already running.")
            return

        # Clean up any leftover failed runs
        self.stop(clean=clean)

        print(f"Starting {self.__class__.__name__}...   ", flush=True, end='')
        kwargs = self._pre_start(**kwargs)

        # This needs to be set in the ChildEngine, either in _start_engine() or at the start of tracking
        self._tracking_initialised = False

        # Resolve input file path before changing cwd
        if input_file:
            if not isinstance(input_file, (str,Path)):
                raise ValueError("`input_file` has to be a string or Path!")
            if self._num_input_files > 1:
                input_file = [FsPath(f).expanduser().resolve() for f in input_file]
                for f in input_file:
                    if not f.exists():
                        raise ValueError(f"Input file {f} does not exist!")
            else:
                input_file = FsPath(input_file).expanduser().resolve()
                if not input_file.exists():
                    raise ValueError(f"Input file {input_file} does not exist!")

        # Set all engine properties that have a setter (this will remove these properties from the kwargs)
        kwargs = self._set_engine_properties(**kwargs)

        # Do some preparations before creating the input file if needed
        kwargs = self._pre_input(**kwargs)

        # Create input file if needed (this will remove the kwargs relevant to the input file and physics)
        kwargs = self._use_input_file(input_file, **kwargs)
        if clean:
            self.clean_input_files(clean_all=False)
        self._preparing_input = False

        # Start the engine in the ChildEngine
        self._start_engine(**kwargs)

        # Done starting
        if self.verbose:
            print(f"{self.__class__.__name__} started.", flush=True)
        else:
            print(f"Done.", flush=True)

    def stop(self, clean=False, **kwargs):
        kwargs = self._stop_engine(**kwargs)
        if clean:
            self.clean(clean_all=True, **kwargs)
        self._restore_engine_properties(clean=clean)
        self._warning_given = False
        self._tracking_initialised = False

    def is_running(self):
        if hasattr(self, '_preparing_input') and self._preparing_input:
            # We need this to allow changing the element settings which otherwise are locked
            return False
        # If we get here, we cannot say if the engine is running or not and we need an
        # implementation in the child class
        return self._is_running()


    def generate_input_file(self, *, clean=True, filename=None, **kwargs):
        # This method manually generates an input file without starting the engine
        if not self._uses_input_file:
            raise ValueError(f"{self.__class__.__name__} does not use input files!")
        if self._element_dict:
            raise ValueError("Elements already assigned to engine (cannot regenerate input "
                           + "file after starting engine)!")

        # Set all engine properties that have a setter (this will remove these properties from the kwargs)
        cwd = kwargs.get('cwd')
        if cwd and FsPath(cwd).resolve() == FsPath.cwd().resolve():
            print("Warning: Cannot use current working directory as input folder.")
            kwargs.pop('cwd')
        kwargs = self._set_engine_properties(**kwargs)

        # Do some preparations before creating the input file if needed
        kwargs = self._pre_input(**kwargs)

        # Create input file
        input_file, _ = self._generate_input_file(**kwargs)
        if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
            # Some engines might create multiple input files (like Fluka)
            input_file = [input_file]
        if filename is None:
            if hasattr(self, '_old_cwd') and self._old_cwd is not None:
                new_input_file = [input_file[0].rename(self._old_cwd / input_file[0].name)]
            else:
                new_input_file = [input_file[0]]
        else:
            new_input_file = [input_file[0].rename(filename)]
        for i, file in enumerate(input_file[1:]):
            new_input_file.append(file.rename(new_input_file[0].parent / file.name))

        if clean:
            self.clean_input_files(clean_all=True, input_file=input_file)
        self._restore_engine_properties(clean=clean)

        return new_input_file


    def assert_particle_ref(self):
        if self.particle_ref is None:
            raise ValueError(f"{self.__class__.__name__} reference particle not set!")

    def assert_ready_to_track_or_skip(self, coll, particles, _necessary_attributes=[]):
        self._assert_element(coll)

        missing_attributes = False
        for attr in _necessary_attributes:
            if not hasattr(coll, attr) or not getattr(coll, attr):
                missing_attributes = True

        if not coll.active or not coll._tracking or not coll.jaw or missing_attributes:
            coll._equivalent_drift.track(particles)
            return False

        npart = particles._num_active_particles
        if npart == 0:
            return False
        if not isinstance(particles._buffer.context, xo.ContextCpu):
            raise ValueError(f"{self.__class__.__name__} only supports CPU contexts!")

        assert self.environment.compiled
        if not self.is_running():
            raise ValueError(f"{self.__class__.__name__} not yet running!\nPlease do this "
                           + f"first, by calling xcoll.{self.__class__.__name__}.start().")

        self.assert_particle_ref()
        if abs(particles.mass0 - self.particle_ref.mass0) > 1e-3:
            raise ValueError(f"Error in reference mass of `particles`: not in sync with "
                           + f"{self.name} reference particle!\nRebuild the particles object "
                           + f"using the {self.__class__.__name__} reference particle.")
        if abs(particles.q0 - self.particle_ref.q0) > 1e-3:
            raise ValueError(f"Error in reference charge of `particles`: not in sync with "
                           + f"{self.name} reference particle!\nRebuild the particles object "
                           + f"using the {self.__class__.__name__} reference particle.")
        if not self._only_protons:
            if np.any([pdg_id == 0 for pdg_id in particles.pdg_id]):
                raise ValueError("Some particles are missing the pdg_id!")
            if particles._num_active_particles + particles._num_lost_particles == particles._capacity:
                raise ValueError("Particles capacity equal to size! Please provide extra capacity "
                               + "for secondaries.")
        return True


    def clean(self, **kwargs):
        self.clean_input_files(**kwargs)
        self.clean_output_files(**kwargs)

    def clean_input_files(self, clean_all=False, **kwargs):
        kwargs = self._get_input_cwd_for_cleaning(**kwargs)
        files_to_delete = self._get_input_files_to_clean(**kwargs)
        if not clean_all:
            files_to_delete = [f for f in files_to_delete
                    if f not in self._all_input_files(kwargs['input_file'])]
        for f in files_to_delete:
            if f is not None and f.exists():
                f.unlink()

    def clean_output_files(self, clean_all=False, **kwargs):
        kwargs = self._get_input_cwd_for_cleaning(**kwargs)
        files_to_delete = self._get_output_files_to_clean(**kwargs)
        if not clean_all:
            files_to_delete = [f for f in files_to_delete
                    if f not in self._all_input_files(kwargs['input_file'])]
        for f in files_to_delete:
            if f is not None and f.exists():
                f.unlink()


    # =======================
    # === Private Methods ===
    # =======================

    # For all the engine fields, they can either be set in advance on the engine,
    # or they can be set when the engine is started. In the latter case, the values
    # are temporary and the original will be restored when the engine is stopped.

    def _set_property(self, prop, kwargs):
        val = kwargs.pop(prop, None)
        if val is not None:
            # We only need to update the property when it is not None
            setattr(self, f'_old_{prop}', getattr(self, prop))
            setattr(self, prop, val)

    def _set_engine_properties(self, **kwargs):
        self._preparing_input = True  # We need this to allow changing the element settings which otherwise are locked
        # We need to set the following properties first as they are needed by the others
        self._set_property('verbose', kwargs)
        self._set_property('line', kwargs)
        # The following properties have a specific logic
        self._use_seed(kwargs.pop('seed', None))
        self._use_particle_ref(kwargs.pop('particle_ref', None))
        self._sync_line_particle_ref()
        self._get_elements(kwargs.pop('elements', None), kwargs.pop('names', None))
        self._set_cwd(kwargs.pop('cwd', None))
        # Now we can set the rest of the properties
        self._set_property('capacity', kwargs)
        return kwargs

    def _restore_engine_properties(self, clean=False):
        self._preparing_input = False
        # Reset particle_ref in the line
        if hasattr(self, '_old_line_particle_ref'):
            self.line.particle_ref = self._old_line_particle_ref
            del self._old_line_particle_ref
        # The following properties have a specific logic
        self._reactivate_elements()
        self._reset_cwd(clean=clean)
        # Reset all other properties
        self_attributes = self.__dict__.copy()
        for kk, vv in self_attributes.items():
            if kk.startswith('_old_'):
                prop = kk[5:]
                setattr(self, prop, vv)
                delattr(self, kk)
        self._input_file = None
        self._element_dict = {}

    def _use_seed(self, seed=None):
        if seed is None:
            if self.seed is None:
                rng = np.random.default_rng()
                if self._int32:
                    self.seed = rng.integers(0, int(2**31)) # 32-bit signed
                else:
                    self.seed = rng.integers(0, int(2**63)) # 64-bit signed
        else:
            self._old_seed = self.seed
            self.seed = seed
        self._print(f"Using seed {self.seed}.")

    def _use_particle_ref(self, particle_ref=None):
        # Prefer: provided particle_ref > existing particle_ref > particle_ref from line
        if particle_ref is not None:
            self._old_particle_ref = self.particle_ref
            self.particle_ref = particle_ref
        elif self.particle_ref is None:
            if self.line is None or not hasattr(self.line, 'particle_ref') \
            or self.line.particle_ref is None:
                raise ValueError("Need to provide either a line with a reference "
                               + "particle, or `particle_ref`.")
            self._old_particle_ref = self.particle_ref
            self.particle_ref = self.line.particle_ref
        self._print(f"Using {pdg.get_name_from_pdg_id(self.particle_ref.pdg_id[0])} "
                  + f"with momentum {self.particle_ref.p0c[0]/1.e9:.1f} GeV.")

    def _sync_line_particle_ref(self):
        if self.line is None:
            return
        if self.line.particle_ref is not None \
        and not xt.line._dicts_equal(self.line.particle_ref.to_dict(),
                                     self.particle_ref.to_dict()):
            self._print("Found different reference particle in line. Temporarily overwritten.")
            val = self.line.particle_ref
            if isinstance(val, xt.line.LineParticleRef):
                val = val._resolved
            self._old_line_particle_ref = val
            self.line.particle_ref = self.particle_ref

    def _get_new_element_name(self):
        name = f"{self.name}_el_{self._element_index}"
        self._element_index += 1
        return name

    def _deactivate_element(self, el):
        self._deactivated_elements[el.name] = [el, el.active or True]
        if hasattr(el, 'active'):
            el.active = False
        self._remove_element(el)

    def _assert_element(self, element):
        if not isinstance(element, self._element_classes):
            raise ValueError(f"Element {element.name} is not a "
                            + ", or a ".join([c.__name__ for c in self._element_classes])
                            + ".")

    def _get_elements(self, elements=None, names=None):
        if elements is not None and (not hasattr(elements, '__iter__') or isinstance(elements, str)):
            elements = [elements]
        if names is not None and (not hasattr(names, '__iter__') or isinstance(names, str)):
            names = [names]
        if self.line is None:
            if elements is None:
                raise ValueError("Need to provide either `line` or `elements`.")
            if names is None:
                names = []
                for ee in elements:
                    if hasattr(ee, 'name') and ee.name:
                        names.append(ee.name)
                    else:
                        name = self._get_new_element_name()
                        names.append(name)
                        ee.name = name
            elif len(names) == len(elements):
                for ee, name in zip(elements, names):
                    if hasattr(ee, 'name'):
                        if ee.name != name:
                            self._print(f"Warning: Element name {ee.name} changed to {name}.")
                            ee.name = name
            else:
                raise ValueError("Length of `elements` and `names` doesn't match.")
        else:
            if elements is not None:
                raise ValueError("Cannot provide both `line` and `elements`.")
            if names is None:
                elements, names = self.line.get_elements_of_type(self._element_classes)
            else:
                elements = [self.line[name] for name in names]
        this_names = []
        this_elements = []
        for ee, name in zip(elements, names):
            self._assert_element(ee)
            if ee.jaw is None:
                self._print(f"Warning: Jaw not set for {name}. Ignoring.")
                self._deactivate_element(ee)
            elif not ee.active:
                self._print(f"Warning: Element {name} is not active. Ignoring.")
                self._deactivate_element(ee)
            else:
                this_names.append(name)
                this_elements.append(ee)
        if len(this_elements) == 0:
            raise ValueError(f"No active {self.name} elements found!")
        if len(set(this_names)) != len(this_names):
            raise ValueError(f"Duplicate names found in {self.name} elements: {this_names}. "
                           + f"Please provide unique names for each element.")
        self._element_dict = dict(zip(this_names, this_elements))

    def _reactivate_elements(self):
        names = list(self._deactivated_elements.keys())
        for name in names:
            ee, was_active = self._deactivated_elements.pop(name)
            self._restore_element(ee)
            if hasattr(ee, 'active'):
                ee.active = was_active

    def _set_cwd(self, cwd=None):
        if self._uses_run_folder:
            if cwd is not None:
                cwd = FsPath(cwd).expanduser().resolve()
                if cwd.exists() and cwd != FsPath.cwd():
                    # Check if the folder already exists
                    # If the specified folder is the current one, we do not need to rename
                    i = 0
                    while (cwd.parent / f'{cwd.name}_{i:0>4}').exists():
                        i += 1
                        if i > 9999:
                            raise ValueError(f"Too many folders with the same "
                                           + f"name {cwd}!")
                    cwd = cwd.parent / f'{cwd.name}_{i:0>4}'
            else:
                ran_str = ranID(only_alphanumeric=True)
                cwd = FsPath.cwd() / f'{self.name}_run_{ran_str}'
            self._cwd = cwd
            cwd.mkdir(parents=True, exist_ok=True)
            self._old_cwd = FsPath.cwd()
            os.chdir(cwd)

    def _reset_cwd(self, clean=False):
        if hasattr(self, '_old_cwd'):
            if self._old_cwd is not None:
                os.chdir(self._old_cwd)
                if clean and self._cwd is not None and self._cwd != FsPath.cwd():
                    try:
                        self._cwd.rmdir()
                    except:
                        self._print(f"Warning: Failed to remove temporary folder "
                                  + f"{self._cwd}.")
            del self._old_cwd
        self._cwd = None


    def _use_input_file(self, input_file=None, **kwargs):
        if self._uses_input_file:
            if input_file is None:
                input_file, kwargs = self._generate_input_file(**kwargs)
            if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
                # Some engines might need multiple input files (like Fluka)
                input_file = [input_file]
            input_file = [FsPath(f).expanduser().resolve() for f in input_file]
            new_files = []
            for file in input_file:
                if not file.exists():
                    raise ValueError(f"Input file {file.as_posix()} not found!")
                if file.parent != FsPath.cwd() and self._uses_run_folder:
                    file.copy_to(FsPath.cwd(), method='mount')
                    new_files.append(FsPath.cwd() / file.name)
                else:
                    new_files.append(file)
            self._input_file = new_files[0] if self._num_input_files==1 else new_files
            self._match_input_file()
        return kwargs


    def _get_input_cwd_for_cleaning(self, **kwargs):
        # Get the input file and the cwd
        input_file = kwargs.get('input_file', None)
        cwd = kwargs.get('cwd', None)
        if input_file is None:
            if self.input_file is not None:
                input_file = self.input_file
            if cwd is None and self._cwd is not None:
                cwd = self._cwd
        else:
            if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
                input_file = [input_file]
            input_file = [FsPath(ff) for ff in input_file]
            if cwd is None:
                cwd = input_file[0].parent
            if len(input_file) == 1:
                input_file = input_file[0]
        kwargs['input_file'] = input_file
        if self._uses_run_folder:
            kwargs['cwd'] = cwd
        else:
            kwargs['cwd'] = FsPath.cwd()
        return kwargs


    # =================================================
    # === Methods to be overwritten by child engine ===
    # =================================================

    def _start_engine(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_start_engine` for "
                                   + f"{self.__class__.__name__}!")

    def _stop_engine(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_stop_engine` for "
                                   + f"{self.__class__.__name__}!")

    def _is_running(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_is_running` for "
                                   + f"{self.__class__.__name__}!")

    def _match_input_file(self, **kwargs):
        if self._uses_input_file:
            raise NotImplementedError(f"Need to implement `_match_input_file` for "
                                   + f"{self.__class__.__name__}!")

    def _generate_input_file(self, **kwargs):
        if self._uses_input_file:
            raise NotImplementedError("Need to implement `_generate_input_file` for "
                                   + f"{self.__class__.__name__}!")


    # ============================================================
    # === Methods to be optionally overwritten by child engine ===
    # ============================================================

    def _get_input_files_to_clean(self, input_file=None, cwd=None, clean_all=False, **kwargs):
        return []

    def _get_output_files_to_clean(self, input_file=None, cwd=None, clean_all=False, **kwargs):
        return []

    def _all_input_files(self, input_file=None):
        if self._uses_input_file:
            if self._num_input_files == 1:
                return [self._input_file]
            else:
                raise NotImplementedError(f"Need to implement `_all_input_files` for "
                                    + f"{self.__class__.__name__}!")
        else:
                return []

    def _remove_element(self, el):
        pass

    def _restore_element(self, el):
        pass

    def _pre_start(self, **kwargs):
        return kwargs

    def _pre_input(self, **kwargs):
        # Do some preparations before creating the input file if needed
        return kwargs



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import CrystalMaterial
from .atoms import Silicon, Germanium, Tungsten
from .allotropes import CarbonFibreCarbon


CarbonCrystal    = CrystalMaterial.from_material(CarbonFibreCarbon, name='CarbonCrystal',
                                    crystal_plane_distance=0.63e-7, eta=0.9,
                                    crystal_potential=21.0, nuclear_collision_length=1.e-12)

SiliconCrystal   = CrystalMaterial.from_material(Silicon, crystal_plane_distance=0.96e-7,
                                    crystal_potential=21.34, nuclear_collision_length=0.3016,
                                    eta=0.9)

GermaniumCrystal = CrystalMaterial.from_material(Germanium, crystal_plane_distance=1.0e-7,
                                    crystal_potential=40.0, nuclear_collision_length=0.1632,
                                    eta=0.9)

TungstenCrystal  = CrystalMaterial.from_material(Tungsten, crystal_plane_distance=0.56e-7,
                                    crystal_potential=21.0, nuclear_collision_length=1.e-12,
                                    eta=0.9)

# Clean up namespace
del CarbonFibreCarbon, Silicon, Germanium, Tungsten
del CrystalMaterial



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import io
import xtrack as xt


class BaseWrapper:
    _engine_cls = None
    _environment_cls = None

    def __init__(self):
        self._engine = None
        self._environment = None

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def engine(self):
        self._lazy_load_engine()
        return self._engine

    @property
    def environment(self):
        self._lazy_load_environment()
        return self._environment

    def _lazy_load_environment(self):
        # Ensure the environment is loaded
        if not self._environment:
            self._environment = self._environment_cls()

    def _lazy_load_engine(self):
        # Ensure the engine is loaded
        if not self._engine:
            self._engine = self._engine_cls()
            self._engine._environment = self.environment



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material
from .database import db
from .atoms import (Carbon, Oxygen, Argon, Hydrogen, Boron, Nitrogen, Silicon,
                    Aluminium, Iron, Nickel, Cobalt, Zinc, Chromium, Titanium,
                    Zirconium, Manganese, Phosphorus, Sulfur, Copper, Molybdenum,
                    Tungsten, Calcium, Niobium, Magnesium, Tantalum)
from .compounds import CarbonDioxide, BoronNitride, Silica, BoronTrioxide, AluminiumOxide, NiZnFerrite


# TODO: add more from Geant4: https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Appendix/materialNames.html


# Mixture definitions from FLUKA
# ===============================

# By molar fractions (like number of atoms but not a compound)
CopperDiamond      = Material(components=[Copper, Carbon, Chromium, Zirconium], molar_fractions=[0.2696, 0.7269, 0.0033, 0.0002], density=5.4, state='solid')
TiZrMo             = Material(components=[Molybdenum, Carbon, Titanium, Zirconium], molar_fractions=[0.98762, 0.00158, 0.00996, 0.00084], density=10.22, state='solid')
MolybdenumGraphite = Material(components=[Molybdenum, Carbon, Titanium], molar_fractions=[0.0201, 0.978, 0.0019], density=2.57, state='solid')
MG6403Fc           = Material(components=[Molybdenum, Carbon, Titanium], molar_fractions=[0.0184, 0.9809, 0.0007], density=2.55, state='solid')
MG6400             = Material(components=[Molybdenum, Carbon], molar_fractions=[0.0146, 0.9854], density=2.48, state='solid')

MolybdenumGraphite.info = 'Molybdenum-Graphite composite (NB-8404Ng), used for LHC TCPPM and TCSPM.'
MG6403Fc.info           = 'Molybdenum-Graphite composite. Candidate material for LHC collimators, but only used for TCPPMP prototype.'
MG6400.info             = 'Molybdenum-Graphite composite. Candidate material for LHC collimators, but not used.'

# By mass fractions
Air                 = Material(components=[Nitrogen, Oxygen, Argon, CarbonDioxide], state='gas', pressure=1, temperature=288.15,
                               mass_fractions=[78.084, 20.946, 0.934, 0.036], density=0.001225)
BoronNitride5000    = Material(components=[BoronNitride, Silica, Oxygen, BoronTrioxide, Calcium], state='solid',
                               mass_fractions=[0.929, 0.002, 0.04, 0.004, 0.025], density=1.925)
FeNiCo              = Material(components=[Iron, Nickel, Cobalt], state='solid',
                               mass_fractions=[0.54, 0.29, 0.17], density=8.35)
Glidcop15           = Material(components=[Copper, AluminiumOxide], state='solid',
                               mass_fractions=[0.9985, 0.0015], density=8.93)
CuNiFeMn            = Material(components=[Copper, Nickel, Iron, Manganese], state='solid',
                               mass_fractions=[0.875, 0.1, 0.015, 0.01], density=8.87)
Inermet180          = Material(components=[Tungsten, Nickel, Copper], state='solid',
                               mass_fractions=[0.95, 0.035, 0.015], density=18.0)
StainLessSteel304L  = Material(components=[Iron, Chromium, Nickel, Manganese, Silicon, Phosphorus, Sulfur, Carbon], state='solid',
                               mass_fractions=[0.67145, 0.185, 0.1125, 0.02, 0.01, 0.00045, 0.0003, 0.0003], density=8.02)
StainLessSteel316L  = Material(components=[Iron, Chromium, Nickel, Molybdenum, Manganese, Silicon, Phosphorus, Sulfur, Carbon], state='solid',
                               mass_fractions=[0.61395, 0.185, 0.14, 0.03, 0.02, 0.01, 0.00045, 0.0003, 0.0003], density=8.0)
StainLessSteel316LN = Material(components=[Iron, Chromium, Nickel, Molybdenum, Manganese, Silicon, Titanium, Nitrogen, Niobium, Copper, Cobalt, Phosphorus, Carbon, Sulfur, Tantalum, Boron], state='solid',
                               mass_fractions=[0.65093, 0.17, 0.12, 0.025, 0.02, 0.0075, 0.0015, 0.0014, 0.001, 0.001, 0.0005, 0.00045, 0.0003, 0.0003, 0.0001, 0.00002], density=8.03)
TitaniumNitride     = Material(components=[Titanium, Nitrogen], state='solid',
                               mass_fractions=[0.77, 0.23], density=5.22)
Antico              = Material(components=[Aluminium, Silicon, Iron, Magnesium, Manganese, Chromium, Titanium, Zinc, Copper], state='solid',
                               mass_fractions=[0.9725, 0.006, 0.005, 0.004, 0.004, 0.0035, 0.002, 0.002, 0.001], density=2.7)
CopperChromium      = Material(components=[Copper, Chromium], state='solid',
                               mass_fractions=[0.991, 0.009], density=8.8)
CastIron            = Material(components=[Iron, Carbon, Silicon], state='solid',
                               mass_fractions=[0.94, 0.04, 0.02], density=7.2)
CuCrZr              = Material(components=[Copper, Chromium, Iron, Silicon, Zirconium], state='solid',
                               mass_fractions=[0.98805, 0.0085, 0.0008, 0.001, 0.00165],  density=8.9)
TiGr2               = Material(components=[Titanium, Iron, Carbon, Oxygen, Nitrogen, Hydrogen], state='solid',
                               mass_fractions=[0.99325, 0.003, 0.0008, 0.0025, 0.0003, 0.00015], density=4.51)


# Extra parameters for Everest
# ============================

MolybdenumGraphite.adapt(inplace=True, nuclear_radius=0.25, nuclear_elastic_slope=76.7,
                         cross_section=[0.362, 0.247, 0, 0, 0, 0.0094e-2], hcut=0.02)
CopperDiamond.adapt(inplace=True, nuclear_radius=0.308, nuclear_elastic_slope=115.0,
                    cross_section=[0.572, 0.370, 0, 0, 0, 0.0279e-2], hcut=0.02)
Glidcop15.adapt(inplace=True, nuclear_radius=0.418, nuclear_elastic_slope=208.7,
                cross_section=[1.246, 0.765, 0, 0, 0, 0.1390e-2], hcut=0.02)
Inermet180.adapt(inplace=True, nuclear_radius=0.578, nuclear_elastic_slope=392.1,
                 cross_section=[2.548, 1.473, 0, 0, 0, 0.5740e-2], hcut=0.02)


# Existing FLUKA mixtures
# =======================

# POLYSTYR Polystyrene 1.06
# PLASCINT Plastic scintillator 1.032
# PMMA Polymethyl methacrylate, Plexiglas, Lucite, Perspex 1.19
# BONECOMP Compact bone 1.85
# BONECORT Cortical bone 1.85
# MUSCLESK Skeletal muscle 1.04
# MUSCLEST Striated muscle 1.04
# ADTISSUE Adipose tissue 0.92
# KAPTON Kapton polyimide film 1.42
# POLYETHY Polyethylene 0.94


# Metadata for database
# =====================

del Carbon, Oxygen, Argon, Hydrogen, Boron, Nitrogen, Silicon, Aluminium, Iron, Nickel, Cobalt, Zinc, Chromium, Titanium
del Zirconium, Manganese, Phosphorus, Sulfur, Copper, Molybdenum, Tungsten, Calcium, Niobium, Magnesium, Tantalum
del CarbonDioxide, BoronNitride, Silica, BoronTrioxide, AluminiumOxide, NiZnFerrite
for name, obj in list(globals().items()):  # Have to wrap in list to take a snapshot (avoid updating in-place)
    if isinstance(obj, Material) and obj.name is None:
        obj.name = name

CopperDiamond.fluka_name = 'CUDIAM75'
MG6403Fc.fluka_name = 'MG6403Fc'
MG6400.fluka_name = 'MG6400'
TiZrMo.fluka_name = 'TZM'
Air.fluka_name = 'AIR'
BoronNitride5000.fluka_name = 'BN5000'
FeNiCo.fluka_name = 'ASTMF-15'
Glidcop15.fluka_name = 'GLIDCP15'
CuNiFeMn.fluka_name = 'CuNiFeMn'
Inermet180.fluka_name = 'INERM180'
StainLessSteel304L.fluka_name = 'SS304L'
StainLessSteel316L.fluka_name = 'SS316L'
StainLessSteel316LN.fluka_name = 'SS316LN'
TitaniumNitride.fluka_name = 'TITNAT'
Antico.fluka_name = 'ANTICO'
CopperChromium.fluka_name = 'CuCr'
CastIron.fluka_name = 'CASTIRON'
CuCrZr.fluka_name = 'CuCrZr'
TiGr2.fluka_name = 'TiGr2'

db['MoGR'] = MolybdenumGraphite
db['CuCD'] = CopperDiamond
db['Glid'] = Glidcop15
db['Iner'] = Inermet180


# Clean up namespace
del name, obj
del Material
del db



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
import xpart as xp
import time
import io

from ...headers.particle_states import (LOST_WITHOUT_SPEC, LOST_ON_GEANT4_COLL,
    EXCITED_ION_STATE, MASSLESS_OR_NEUTRAL, VIRTUAL_ENERGY, HIT_ON_GEANT4_COLL)


# def _drift(coll, particles, length):
#     old_length = coll._equivalent_drift.length
#     coll._equivalent_drift.length = length
#     coll._equivalent_drift.track(particles)
#     coll._equivalent_drift.length = old_length

def track_pre(coll, particles):
    import xcoll as xc

    # Initialize ionisation loss accumulation variable
    if coll._acc_ionisation_loss < 0:
        coll._acc_ionisation_loss = 0.

    if not xc.geant4.engine.assert_ready_to_track_or_skip(
        coll, particles, _necessary_attributes=['geant4_id']):
        return False  # Stop tracking

    return True  # Continue tracking


def track_post(coll, particles):
    alive_states = np.unique(particles.state[particles.state > 0])
    assert len(alive_states) <= 1, f"Unexpected alive particle states after tracking: {alive_states}"
    if len(alive_states) == 1:
        assert alive_states[0] == 1, f"Unexpected alive particle state after tracking: {alive_states[0]}"


# TODO: need to rework this logic with HIT_ON_GEANT4_COLL


def track_core(coll, part):
    import xcoll as xc
    xc.geant4.engine._g4link.clearData() # Clear the old data - bunch particles and hits

    # send_to_geant4 = part.state == HIT_ON_GEANT4_COLL
    send_to_geant4 = part.state < 1.e21 # Temporary: send all particles to Geant4
    # npart          = send_to_geant4.sum()
    # max_id         = part.particle_id[part.state > -9999].max()
    # assert npart  <= part._num_active_particles
    # if npart == 0:
    #     return

    # This temp delta is necessary because for primary particles, the coordinates are
    # modified in place. But for the longitudinal plane there are 3 coordinates that must
    # be updated, so pass a copy of the delta for the update in place and trigger the
    # correct update of the 3 coordinates later
    delta_temp = part._delta[send_to_geant4].copy()
    npart = part._num_active_particles
    ndead = part._num_lost_particles

    if xc.geant4.engine.reentry_protection_enabled:
        ### remove this part after geant4 bug fixed
        coords = {
            'x': part.x[send_to_geant4],
            'y': part.y[send_to_geant4],
            'px': part.px[send_to_geant4],
            'py': part.py[send_to_geant4],
            'zeta': part.zeta[send_to_geant4],
            'delta': delta_temp,
            'chi': part.chi[send_to_geant4],
            'charge_ratio': part.charge_ratio[send_to_geant4],
            's': part.s[send_to_geant4],
            'pdg_id': part.pdg_id[send_to_geant4],
            'particle_id': part.particle_id[send_to_geant4],
            'state': part.state[send_to_geant4],     # TODO:  does collimasim work if states are HIT_ON_GEANT4_COLL?
            'at_element': part.at_element[send_to_geant4],
            'at_turn': part.at_turn[send_to_geant4]
        }
        buf = io.BytesIO() # Use numpy.savez to serialize
        np.savez(buf, **coords)
        buf.seek(0)
        result_blob = xc.geant4.engine._g4link.add_particles_and_collimate_return(
                                                buf.getvalue(), f'{coll.geant4_id}') # TODO: should geant4_id be a string or an int?
        result_buf = io.BytesIO(result_blob) # Deserialize
        products = np.load(result_buf)
    else:
        coords = [part.x[send_to_geant4], part.y[send_to_geant4], part.px[send_to_geant4], part.py[send_to_geant4],
                  part.zeta[send_to_geant4], delta_temp, part.chi[send_to_geant4],
                  part.charge_ratio[send_to_geant4], part.s[send_to_geant4],
                  part.pdg_id[send_to_geant4], part.particle_id[send_to_geant4], part.state[send_to_geant4],
                  part.at_element[send_to_geant4], part.at_turn[send_to_geant4]]
        xc.geant4.engine._g4link.addParticles(coords)
        xc.geant4.engine._g4link.selectCollimator(f'{coll.geant4_id}')  # TODO: should geant4_id be a string or an int?
        xc.geant4.engine._g4link.collimate()
        products = xc.geant4.engine._g4link.collimateReturn(coords)

    npartsAliveAndDead = npart+ndead
    part.x[:npartsAliveAndDead] = products['x'][:npartsAliveAndDead]
    part.y[:npartsAliveAndDead] = products['y'][:npartsAliveAndDead]
    part.px[:npartsAliveAndDead] = products['px'][:npartsAliveAndDead]
    part.py[:npartsAliveAndDead] = products['py'][:npartsAliveAndDead]
    part.zeta[:npartsAliveAndDead] = products['zeta'][:npartsAliveAndDead]
    part.charge_ratio[:npartsAliveAndDead] = products['charge_ratio'][:npartsAliveAndDead]
    part.s[:npartsAliveAndDead] = products['s'][:npartsAliveAndDead]
    part.pdg_id[:npartsAliveAndDead] = products['pdg_id'][:npartsAliveAndDead]
    # parent particle id should NOT be updated for particles that were sent in
    part.at_element[:npartsAliveAndDead] = products['at_element'][:npartsAliveAndDead]
    part.at_turn[:npartsAliveAndDead] = products['at_turn'][:npartsAliveAndDead]
    part.state[:npartsAliveAndDead] = products['state'][:npartsAliveAndDead]
    new_delta = part.delta.copy()
    new_delta[:npartsAliveAndDead] = products['delta'][:npartsAliveAndDead]
    part.update_delta(new_delta)
    if products['x'] is None or products['x'][npartsAliveAndDead] == -9999:
        part.reorganize()
    else:
        mask = products['state'][npartsAliveAndDead:] > -999999
        new_particles = xp.Particles(_context=part._buffer.context,
                p0c = part.p0c[0],
                mass0 = part.mass0,
                q0 = part.q0,
                s = products['s'][npartsAliveAndDead:][mask],
                x = products['x'][npartsAliveAndDead:][mask],
                px = products['px'][npartsAliveAndDead:][mask],
                y = products['y'][npartsAliveAndDead:][mask],
                py = products['py'][npartsAliveAndDead:][mask],
                zeta = products['zeta'][npartsAliveAndDead:][mask],
                delta = products['delta'][npartsAliveAndDead:][mask],
                mass_ratio = products['mass_ratio'][npartsAliveAndDead:][mask],
                charge_ratio = products['charge_ratio'][npartsAliveAndDead:][mask],
                at_element = products['at_element'][npartsAliveAndDead:][mask],
                at_turn = products['at_turn'][npartsAliveAndDead:][mask],
                parent_particle_id = products['parent_particle_id'][npartsAliveAndDead:][mask],
                pdg_id = products['pdg_id'][npartsAliveAndDead:][mask])

        part.add_particles(new_particles)

    # Set the state of excited ions - not supported (will fail in BDSIM when resent)
    mask = (part.pdg_id > 999999999) & ((part.pdg_id % 10) != 0)
    part.state[mask] = EXCITED_ION_STATE

    # Set the dead state
    mask = part.state == LOST_WITHOUT_SPEC
    part.state[mask] = LOST_ON_GEANT4_COLL



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025                  #
# ######################################### #

import os
import sys
from contextlib import contextmanager


@contextmanager
def pin_python_stdio():
    ORIG_STDOUT_FD = os.dup(1)
    ORIG_STDERR_FD = os.dup(2)
    try:
        yield

    finally:
        # flush the duplicates
        try: sys.stdout.flush()
        except Exception: pass
        try: sys.stderr.flush()
        except Exception: pass

        # restore the originals *before* closing the duplicates
        sys.stdout = os.fdopen(os.dup(ORIG_STDOUT_FD), "w", buffering=1)
        sys.stderr = os.fdopen(os.dup(ORIG_STDERR_FD), "w", buffering=1)

        # now it's safe to close the duplicates
        try: ORIG_STDOUT_FD.close()
        except Exception: pass
        try:
            if ORIG_STDERR_FD is not ORIG_STDOUT_FD:
                ORIG_STDERR_FD.close()
        except Exception: pass



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from contextlib import contextmanager

import xobjects as xo
import xtrack as xt

from .base import BaseCollimator, BaseCrystal
from ..general import _pkg_root
from ..scattering_routines.geant4 import Geant4Engine, track_pre, track_core, track_post
from ..materials import(Material, CrystalMaterial, RefMaterial, db as material_db,
                        _DEFAULT_MATERIAL, _DEFAULT_CRYSTALMATERIAL)


class Geant4Collimator(BaseCollimator):
    _xofields = BaseCollimator._xofields | {
        'geant4_id': xo.String,
        '_tracking': xo.Int8,
        '_acc_ionisation_loss':  xo.Float64,  # TODO: this is not very robust, for when a track is done with new particles etc
    }

    isthick = True
    allow_track = True
    iscollective = True
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _depends_on = [BaseCollimator, Geant4Engine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','geant4_collimator.h')
    ]

    _noexpr_fields         = {*BaseCollimator._noexpr_fields, 'material'}
    _skip_in_to_dict       = [*BaseCollimator._skip_in_to_dict, '_material']
    _store_in_to_dict      = [*BaseCollimator._store_in_to_dict, 'material']
    _internal_record_class = BaseCollimator._internal_record_class

    _allowed_fields_when_frozen = ['_tracking', '_acc_ionisation_loss']

    def __new__(cls, *args, **kwargs):
        with cls._in_constructor():
            self = super().__new__(cls, *args, **kwargs)
        return self

    def __init__(self, **kwargs):
        import xcoll as xc
        with self.__class__._in_constructor(self):
            to_assign = {}
            if '_xobject' not in kwargs:
                kwargs.setdefault('_tracking', True)
                kwargs.setdefault('_acc_ionisation_loss', -1.)
                kwargs.setdefault('geant4_id', ''.ljust(16))
                to_assign['name'] = xc.geant4.engine._get_new_element_name()
                to_assign['material'] = kwargs.pop('material', None)
                kwargs['_material'] = _DEFAULT_MATERIAL
            super().__init__(**kwargs)
            for key, val in to_assign.items():
                setattr(self, key, val)
            if not hasattr(self, '_equivalent_drift'):
                self._equivalent_drift = xt.Drift(length=self.length)

    @property
    def angle(self):
        return BaseCollimator.angle.fget(self)

    @angle.setter
    def angle(self, val):
        if hasattr(val, '__iter__'):
            raise ValueError('The Geant4 scattering engine does not '
                           + 'support unequal jaw rotation angles')
        BaseCollimator.angle.fset(self, val)


    @property
    def material(self):
        if self._material != _DEFAULT_MATERIAL:
            return self._material

    @material.setter
    def material(self, material):
        if material is None:
            material = _DEFAULT_MATERIAL
        elif isinstance(material, dict):
            material = Material.from_dict(material)
        elif isinstance(material, str):
            material = material_db[material]
        elif isinstance(material, RefMaterial):
            if material.geant4_name is None:
                raise ValueError(f"RefMaterial {material} does not have a Geant4 name!")
        if not isinstance(material, Material) \
        or isinstance(material, CrystalMaterial):
            raise ValueError(f"Invalid material of type {type(material)}!")
        if self.material != material:
            self._material = material


    def track(self, part):
        if track_pre(self, part):
            # super().track(part)
            track_core(self, part)
            track_post(self, part)


    def __setattr__(self, name, value):
        import xcoll as xc
        if name not in self._allowed_fields_when_frozen \
        and xc.geant4.engine.is_running():
            raise ValueError('Engine is running; Geant4Collimator is frozen.')
        super().__setattr__(name, value)

    @classmethod
    @contextmanager
    def _in_constructor(cls, self=None):
        original_setattr = cls.__setattr__
        if self is not None:
            self._being_constructed_ = True
        def new_setattr(self, *args, **kwargs):
            return super().__setattr__( *args, **kwargs)
        cls.__setattr__ = new_setattr
        try:
            yield
        finally:
            cls.__setattr__ = original_setattr
            if self is not None:
                self._being_constructed_ = False

    def _being_constructed(self):
        if hasattr(self, '_being_constructed_'):
            return self._being_constructed_
        else:
            return False


class Geant4Crystal(BaseCrystal):
    def __init__(self, **kwargs):
        raise NotImplementedError("Geant4Crystal not yet implemented.")



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material
from .atoms import Carbon, Oxygen, Hydrogen, Boron, Nitrogen, Silicon, Aluminium, Iron, Nickel, Zinc
from .database import db


# General compounds
# =================

CarbonDioxide = Material(components=[Carbon, Oxygen],   n_atoms=[1,2], density=1.951e-3, state='gas', temperature=273.15, pressure=1)
DryIce        = Material(components=[Carbon, Oxygen],   n_atoms=[1,2], density=1.564,    state='solid')
Water         = Material(components=[Hydrogen, Oxygen], n_atoms=[2,1], density=1.0,      state='liquid')


# Compound definitions from FLUKA
# ===============================

BoronNitride   = Material(components=[Boron, Nitrogen],            n_atoms=[1, 1],             density=2.25, state='solid')
Silica         = Material(components=[Silicon, Oxygen],            n_atoms=[1, 2],             density=1.92, state='solid')
BoronTrioxide  = Material(components=[Boron, Oxygen],              n_atoms=[2, 3],             density=2.46, state='solid')
AluminiumOxide = Material(components=[Aluminium, Oxygen],          n_atoms=[2, 3],             density=3.98, state='solid')
NiZnFerrite    = Material(components=[Nickel, Zinc, Iron, Oxygen], n_atoms=[0.45, 0.55, 2, 4], density=5.2,  state='solid')


# Metadata for database
# =====================

del Carbon, Oxygen, Hydrogen, Boron, Nitrogen, Silicon, Aluminium, Iron, Nickel, Zinc
for name, obj in list(globals().items()):  # Have to wrap in list to take a snapshot (avoid updating in-place)
    if isinstance(obj, Material) and obj.name is None:
        obj.name = name

db['Sand'] = Silica

Water.fluka_name = 'WATER'
BoronNitride.fluka_name = 'hBNpure'
Silica.fluka_name = 'SiO2'
BoronTrioxide.fluka_name = 'B2O3'
AluminiumOxide.fluka_name = 'Al2O3'
NiZnFerrite.fluka_name = 'NIZNFER2'


# Clean up namespace
del name, obj
del Material
del db
